' demo program for classification and prediction using adaptive boosting
' James McCaffrey

Namespace AdaptiveBoosting
  Friend Class AdaptiveBoostingProgram
        Shared Sub Main(ByVal args() As String)
            Try
                Console.WriteLine(vbCrLf & "Begin adaptive boosting classification demo" & vbCrLf)

                Dim features() As String = {"Opponent", "Field", "Spread", "Result"} ' dependent feature to predict is last listed

                Dim values(3)() As String
                values(0) = New String() {"Atlanta", "Buffalo", "Chicago", "Detroit"} ' opponent
                values(1) = New String() {"Home", "Away"} ' field
                values(2) = New String() {"Small ", "Medium", "Large "} ' point spread (note: some spaces added)
                values(3) = New String() {"Lose", "Win"} ' the dependent/predicted variable

                Dim rawTrain(9)() As String ' typically training data would be in a text file or SQL table
                rawTrain(0) = New String() {"Detroit", "Home", "Large ", "Win"}
                rawTrain(1) = New String() {"Detroit", "Away", "Medium", "Win"}
                rawTrain(2) = New String() {"Buffalo", "Home", "Small ", "Win"}
                rawTrain(3) = New String() {"Buffalo", "Home", "Medium", "Win"}
                rawTrain(4) = New String() {"Atlanta", "Away", "Large ", "Win"}
                rawTrain(5) = New String() {"Chicago", "Home", "Medium", "Win"}

                rawTrain(6) = New String() {"Chicago", "Away", "Small ", "Lose"}
                rawTrain(7) = New String() {"Chicago", "Home", "Small ", "Lose"}
                rawTrain(8) = New String() {"Atlanta", "Away", "Medium", "Lose"}
                rawTrain(9) = New String() {"Detroit", "Away", "Large ", "Lose"}

                Console.WriteLine("Raw (string) training data for team Seattle:" & vbCrLf)
                Console.WriteLine("Opponent Field  Spread   Result")
                Console.WriteLine("===============================")
                ShowMatrix(rawTrain)

                Console.WriteLine(vbCrLf & "Converting and storing training data.")
                Dim train()() As Integer = RawTrainToInt(rawTrain, values) ' training tuples stored as ints in a matrix; last col (Y) as -1 or +1
                Console.WriteLine("Training data in int form:" & vbCrLf)
                ShowMatrix(train, True)

                Console.WriteLine(vbCrLf & "Creating weak categorical stump learners from training data.")
                Dim learners As List(Of Learner) = MakeLearners(values, train)
                Console.WriteLine("Completed. Weak learners are:" & vbCrLf)
                For i = 0 To learners.Count - 1
                    Console.WriteLine("[" & i & "] " & Description(learners(i), features, values))
                Next i

                Console.WriteLine(vbCrLf & "Initializing list of best learner indexes.")
                Dim bestLearners As New List(Of Integer) ' indexes of best weak learner found for each iteration t in main algorithm loop

                Console.WriteLine(vbCrLf & "Using adaptive boosting to find best learners and their alpha values.")
                MakeModel(train, values, learners, bestLearners) ' find best learners and their alphas (importance)

                Console.WriteLine(vbCrLf & "Model completed.")
                Dim numGood As Integer = bestLearners.Count
                Console.WriteLine("Algorithm found " & numGood & " good learners and associated alpha values.") ' some weak learners may not make the cut

                Console.WriteLine(vbCrLf & "The good learners and their alpha value are:")
                For i = 0 To bestLearners.Count - 1
                    Dim lrn As Integer = bestLearners(i)
                    Console.Write("[" & lrn & "] " & learners(lrn).alpha.ToString("F2") & "  ")
                Next i

                'Console.WriteLine("\nPredicting outcome when Opponent = Buffalo, Field = Away, Spread = Small\n");
                'int] unknownTuple = new int[ { 1, 1, 0 };    // Buffalo, Away, Small

                Console.WriteLine(vbCrLf & "Predicting outcome when Opponent = Detroit, Field = Home, Spread = Small." & vbCrLf)
                Dim unknownTuple() As Integer = {3, 0, 0} ' Detroit, Home, Small

                Dim Y As Integer = Classify(unknownTuple, learners, bestLearners)
                Console.WriteLine("Predicted Y = " & Y & " => Seattle will " & YValueToString(Y, values))

                Console.WriteLine(vbCrLf & "End" & vbCrLf)
                Console.ReadLine()
            Catch ex As Exception
                Console.WriteLine(ex.Message)
                Console.ReadLine()
            End Try
        End Sub ' Main

	' data conversion routines below

        Private Shared Function RawTrainToInt(ByVal rawTrain()() As String, ByVal values()() As String) As Integer()()
            ' converts raw training data in string form to 0-based ints; last col to -1 or +1
            ' calls helper ValueToInt
            Dim rows As Integer = rawTrain.Length
            Dim cols As Integer = rawTrain(0).Length
            Dim result(rows - 1)() As Integer
            For i = 0 To rows - 1
                result(i) = New Integer(cols - 1) {}
            Next i

            For i = 0 To rawTrain.Length - 1
                For j = 0 To rawTrain(i).Length - 1 ' all cols
                    Dim feature As Integer = j
                    Dim v As Integer = ValueToInt(rawTrain(i)(j), feature, values) ' string to int
                    result(i)(j) = v
                Next j
            Next i

            For i = 0 To result.Length - 1 ' convert values in last col from 0,1 to -1,+1
                If result(i)(cols - 1) = 0 Then
                    result(i)(cols - 1) = -1
                End If
            Next i

            Return result

        End Function ' RawTrainToInt

        Private Shared Function ValueToInt(ByVal value As String, ByVal feature As Integer, ByVal values()() As String) As Integer
            ' convert a string value like "Detroit" to an int (3)
            Dim row As Integer = feature
            For j = 0 To values(row).Length - 1
                If values(row)(j) = value Then
                    Return j
                End If
            Next j
            Return -1 ' error
        End Function

        Private Shared Function IntToValue(ByVal value As Integer, ByVal feature As Integer, ByVal values()() As String) As String
            ' convert an int value like 2 to a string ("Chicago")
            Return values(feature)(value)
        End Function

        Private Shared Function YValueToString(ByVal Y As Integer, ByVal values()() As String) As String
            ' convert -1 or +1 to a string like "Lose" or "Win"
            If Y = -1 Then
                Return values(values.Length - 1)(0)
            ElseIf Y = 1 Then
                Return values(values.Length - 1)(1)
            Else
                Return "error"
            End If
        End Function

	Private Shared Function FeatureToInt(ByVal feature As String, ByVal features() As String) As Integer
	  ' convert a string feature like "Opponent" to an int (0)
	  For i  = 0 To features.Length - 1
		If features(i) = feature Then
		  Return i
		End If
	  Next i
	  Return -1 ' error
	End Function

	Private Shared Function IntToFeature(ByVal feature As Integer, ByVal features() As String) As String
	  ' convert an int like 1 to a string ("Field")
	  Return features(feature)
	End Function

	' learner-making routines below. see class Learner too.

        Private Shared Function Description(ByVal learner As Learner, ByVal features() As String, ByVal values()() As String) As String
            ' return a friendly string description of a weak Learner
            Dim feature As String = IntToFeature(learner.feature, features) ' like "Opponent"
            Dim value As String = IntToValue(learner.value, learner.feature, values) ' like "Atlanta"
            Dim dependent As String = features(features.Length - 1)
            Dim predicted As String = ""
            If learner.predicted = 1 Then
                predicted = values(values.Length - 1)(1)
            Else ' predicted = -1
                predicted = values(values.Length - 1)(0)
            End If

            Return "IF " & feature & " IS " & value & " THEN " & dependent & " IS " & predicted & "   (raw errorvar = " & learner.errorVal.ToString("F2") & ")"
        End Function

        Private Shared Function MakeLearners(ByVal values()() As String, ByVal train()() As Integer) As List(Of Learner)
            ' scan through training set and try to make one weak learner for each possible feature value
            ' do not create a learner for values with raw error rate of 0.50 or greater
            Dim numFeatures As Integer = values.Length ' includes the dependent variable
            Dim result As New List(Of Learner)

            For f = 0 To numFeatures - 2 ' each predictive feature (but not Y)
                Dim numValues As Integer = values(f).Length ' num values for curr feature
                For v = 0 To numValues - 1 ' each value of curr feature
                    Dim pred As Integer = MostLikely(f, v, train) ' most likely Y (-1,+1) for v; 0 if no most likely can be determined or error >= 0.50
                    If pred = 0 Then ' 0 means that we could not make a prediction for this value, so we won't make a Learner for this feature value
                        Continue For
                    End If
                    Dim lrn As New Learner(f, v, pred, -1.0, -1.0, -1.0) ' tmp error -1.0, tmp epsilon -1.0, tmp alpha = -1.0
                    result.Add(lrn)
                Next v
            Next f
            ' for each learner, scan training data and compute raw error rate
            For i = 0 To result.Count - 1 ' each learner
                Dim nw As Integer = 0 ' number wrong
                Dim nr As Integer = 0 ' number relevant
                For k = 0 To train.Length - 1 ' each training tuple
                    If IsApplicable(result(i), train(k)) = False Then ' the curr rule doesn't apply to curr tuple so skip
                        Continue For
                    End If
                    nr = nr + 1
                    Dim act As Integer = train(k)(train(k).Length - 1) ' actual -1 or +1 is value in last column of training matrix
                    If result(i).predicted <> act Then ' wrong
                        nw = nw + 1
                    End If
                Next k
                If nr = 0 Then ' consider throWing an Exception here
                    result(i).errorVal = 0.0
                Else
                    result(i).errorVal = (nw * 1.0) / nr
                End If
            Next i

            Return result
        End Function

        Private Shared Function MostLikely(ByVal feature As Integer, ByVal value As Integer, ByVal train()() As Integer) As Integer
            ' highest probability Y value for a particular feature value (like "field" is "Home")
            ' return -1 or +1 based on highest prob,
            ' return 0 if equal percentages (0.5 and 0.5) or no relevant training data
            ' used to determine the predicted Y field in a Learner object in method MakeLearners
            Dim ctMinusOne As Integer = 0 ' relevant count of -1 values: number training tuples where value has Y = -1
            Dim ctPlusOne As Integer = 0 ' relevant count +1 values
            Dim totRelevant As Integer = 0
            Dim cols As Integer = train(0).Length

            For i = 0 To train.Length - 1 ' each training tuple
                If train(i)(feature) <> value Then ' this tuple not relevant because wrong value
                    Continue For
                End If

                totRelevant = totRelevant + 1 ' we know curr tuple is relevant

                If train(i)(cols - 1) = -1 Then ' the last column is the dependent variable (like sex)
                    ctMinusOne = ctMinusOne + 1
                Else
                    ctPlusOne = ctPlusOne + 1
                End If
            Next i

            If totRelevant = 0 Then ' there were no relevant training tuples for the feature value so cannot determine a most likely Y
                Return 0
            End If
            If ctMinusOne > 0 AndAlso ctPlusOne = 0 Then ' there was at least one -1 but no +1 values so most likely Y is -1
                Return -1
            End If
            If ctMinusOne = 0 AndAlso ctPlusOne > 0 Then ' see comment above
                Return 1
            End If

            Dim pctMinusOne As Double = (ctMinusOne * 1.0) / totRelevant ' percentage or probability that Y is -1
            Dim pctPlusOne As Double = (ctPlusOne * 1.0) / totRelevant ' percentage or probability that Y is +1

            If VeryCLose(pctMinusOne, pctPlusOne) = True Then ' both pcts are 0.5000 so no most likely Y
                Return 0
            End If

            If pctMinusOne > pctPlusOne Then
                Return -1
            Else
                Return 1
            End If
        End Function

	' model-building routines below

        Private Shared Sub UpdateEpsilons(ByVal learners As List(Of Learner), ByVal train()() As Integer, ByVal trainWeights() As Double)
            ' for each learner, update its epsilon by walking thru each training tuple and computing a weighted error sum
            For i = 0 To learners.Count - 1
                Dim ep As Double = 0.0
                For k = 0 To train.Length - 1 ' each training tuple
                    If IsApplicable(learners(i), train(k)) = False Then ' curr learner doesn't apply to curr tuple so move to next tuple
                        Continue For
                    End If
                    Dim actual As Integer = train(k)(train(k).Length - 1) ' actual -1 +1 is in last col of training matrix
                    Dim pred As Integer = learners(i).predicted ' predicted -1 +1
                    If actual <> pred Then ' incorrect prediction so..
                        ep = ep + trainWeights(k) ' accumulate weighted error
                    End If
                Next k
                learners(i).epsilon = ep ' could be zero -- must deal with before computing alpha
            Next i ' each learner
        End Sub

	Private Shared Function IsApplicable(ByVal learner As Learner, ByVal tuple() As Integer) As Boolean
	  ' is the learner/weak classifer applicable to the training tuple?
	  Dim ti As Integer = learner.feature ' index into the tuple
	  If tuple(ti) = learner.value Then
		  Return True
		  Else
			  Return False
		  End If
	End Function

	Private Shared Function BestLearner(ByVal learners As List(Of Learner), <System.Runtime.InteropServices.Out> ByRef epsilon As Double, ByVal usedLearners() As Integer) As Integer
	  ' find index of not-yet-used learner that has Smallest epsilon (weighted error)
	  ' update usedLearners hash table array

	  ' check to make sure there is at least one unused weak learner
	  Dim x As Integer = Array.IndexOf(usedLearners, 0) ' find first unused learner
	  If x < 0 Then
		  Throw New Exception("No unused weak learners found by method BestLearner")
	  End If

	  Dim SmallestEpsilon As Double = Double.MaxValue ' Smaller epsilon values mean Smaller error which means better learner
	  Dim bestIndex As Integer = -1
	  For i  = 0 To learners.Count - 1
		If usedLearners(i) = 1 Then ' curr weak learner has already been used
			Continue For
		End If

		If learners(i).epsilon < SmallestEpsilon Then
		  SmallestEpsilon = learners(i).epsilon
		  bestIndex = i
		End If
	  Next i

	  If SmallestEpsilon < 0.0 Then
		  Throw New Exception("Encountered a negative epsilon in method BestLearner")
	  End If
	  If bestIndex = -1 Then
		  Throw New Exception("Unable to find best learner index in method BestLearner")
	  End If
	  epsilon = SmallestEpsilon ' assign out parameter value. could be 0.0
	  usedLearners(bestIndex) = 1 ' flag the best weak learner as used
	  Return bestIndex
	End Function

	Private Shared Function CLoseToZero(ByVal x As Double) As Boolean
	  If Math.Abs(x) < 0.00000001 Then
		  Return True
	  Else
		  Return False
	  End If
	End Function

	Private Shared Function VeryCLose(ByVal x As Double, ByVal y As Double) As Boolean
	  If Math.Abs(x - y) < 0.00000001 Then
		  Return True
	  Else
		  Return False
	  End If
	End Function

        Private Shared Sub UpdateTrainingWeights(ByVal trainWeights() As Double, ByVal train()() As Integer, ByVal learners As List(Of Learner), ByVal alphaT As Double, ByVal bestIndex As Integer)
            ' increase weights of misclassified (by curr best learner) training tuples, decrease weights of correctly classified tuples
            Dim cols As Integer = train(0).Length ' number columns of training data matrix

            For i = 0 To trainWeights.Length - 1 ' each weight and each training tuple
                If IsApplicable(learners(bestIndex), train(i)) = False Then
                    Continue For
                End If
                Dim actual As Integer = train(i)(cols - 1) ' -1 or +1
                trainWeights(i) = trainWeights(i) * Math.Exp(-alphaT * actual * learners(bestIndex).predicted) ' tricky; depends on -1 +1 encoding for Y
            Next i

            Dim Z As Double = 0.0 ' sum of all D weights to normalize them
            For i = 0 To trainWeights.Length - 1 ' each updated weight
                Z = Z + trainWeights(i)
            Next i

            For i = 0 To trainWeights.Length - 1 ' each updated weight
                trainWeights(i) = trainWeights(i) / Z ' normalize (so all weights now sum to 1.0)
            Next i
        End Sub

        Private Shared Sub MakeModel(ByVal train()() As Integer, ByVal values()() As String, ByVal learners As List(Of Learner), ByVal bestLearners As List(Of Integer))
            ' compute (indexes of) best learners and their alphas
            ' assumes all weak learners have been created
            ' assumes bestLearners (index List) has been initialized to a empty List

            Console.WriteLine(vbCrLf & "Initializing training tuple weights ('D') array.")
            Dim trainWeights(train.Length - 1) As Double ' training weights (D values) are local because they aren't needed to classify
            For i = 0 To trainWeights.Length - 1 ' one weight per tuple
                trainWeights(i) = 1.0 / trainWeights.Length ' initially all weights equal
            Next i

            'ShowVector(trainWeights);

            Dim usedLearners(learners.Count - 1) As Integer ' track which weak learners have been added to bestLearners List. index = which learner, value 0 = not used, 1 = used
            Console.WriteLine("Entering main algorithm loop.")
            Dim done As Boolean = False
            Dim t As Integer = 0 ' not sure why research literature uses t for counter variable name here...
            Do While done = False
                'Console.WriteLine("================");
                'Console.WriteLine("t = " + t);
                UpdateEpsilons(learners, train, trainWeights) ' compute weighted error for each learner
                'for (int j = 0; j < learners.Count; ++j)
                '  Console.WriteLine(learnersj.ToString);

                Dim epsilonT As Double ' best (Smallest) weighted error at curr iteration t
                Dim bestIndex As Integer = BestLearner(learners, epsilonT, usedLearners) ' find index of best learner (that hasn't yet been used) and its weighted error epsilonT
                'Console.WriteLine("best learner is " + bestIndex); Console.ReadLine;
                If Math.Abs(0.5 - epsilonT) <= 0.00001 Then ' if best weighted error is (almost) 0.5 or greater then no use in continuing on
                    Exit Do
                End If
                bestLearners.Add(bestIndex)

                If CLoseToZero(epsilonT) = True Then ' avoid division by 0 when computing alphaT
                    epsilonT = 0.000001
                End If
                Dim alphaT As Double = 0.5 * Math.Log((1.0 - epsilonT) / epsilonT) ' Smaller epsilon error makes Larger alpha values
                'Console.WriteLine("a = " & alphaT); Console.ReadLine;
                learners(bestIndex).alpha = alphaT ' store the alpha

                UpdateTrainingWeights(trainWeights, train, learners, alphaT, bestIndex) ' update weights of training tuples that are relevant to best learner

                t = t + 1
                If t = learners.Count Then ' every weak learner has been used
                    done = True
                End If
                'Console.WriteLine("================");
            Loop
            bestLearners.Sort() ' sort this List of indexes to make the output easier to read. not absolutely necessary
        End Sub

	Private Shared Function Classify(ByVal testTuple() As Integer, ByVal learners As List(Of Learner), ByVal bestLearners As List(Of Integer)) As Integer
	  ' use adaptive boosting model (set of weak learners and their alpha (importtance) weights) to predict Y for a testTuple (in int form)
	  Dim accVote As Double = 0.0 ' accumulated votes of learners
	  If bestLearners.Count = 0 Then
		  Throw New Exception("No learners found in List of bestLearners")
	  End If

	  For t  = 0 To bestLearners.Count - 1
		Dim idx As Integer = bestLearners(t) ' index of curr learner
		If IsApplicable(learners(idx), testTuple) = False Then ' curr learner can't vote on the tuple
			Continue For
		End If

		'double vote = alphast * learnersidx.predicted;
		Dim vote As Double = learners(idx).predicted * learners(idx).alpha
		accVote += vote
		Dim sign As String = ""
		If learners(idx).predicted = 1 Then
			sign = "+"
		End If
		'Console.WriteLine("Using learner " + idx + " with alpha = " + alphast.ToString("F2") + " prediction is " + sign + learnersidx.predicted + " so vote = " + vote.ToString("F2") );
		Console.WriteLine("Using learner " & idx & " with alpha = " & learners(idx).alpha.ToString("F2") & " prediction is " & sign & learners(idx).predicted & " so vote = " & vote.ToString("F2"))
	  Next t

	  Console.WriteLine(vbCrLf & "Final accumulated vote is " & accVote.ToString("F2"))
	  If accVote > 0.0 Then
		  Return 1
	  ElseIf accVote < 0.0 Then
		  Return -1
	  Else
		  Return 0
	  End If
	End Function

	' display routines below

	Private Shared Sub ShowVector(ByVal vector() As String)
	  For i  = 0 To vector.Length - 1
		Console.Write(vector(i) & " ")
	  Next i
	  Console.WriteLine("")
	End Sub

	Private Shared Sub ShowVector(ByVal vector() As Integer)
	  For i  = 0 To vector.Length - 1
		Console.Write(vector(i) & " ")
	  Next i
	  Console.WriteLine("")
	End Sub

	Private Shared Sub ShowVector(ByVal vector() As Double)
	  For i  = 0 To vector.Length - 1
		Console.Write(vector(i).ToString("F4") & " ")
	  Next i
	  Console.WriteLine("")
	End Sub

        Private Shared Sub ShowMatrix(ByVal matrix()() As String)
            For i = 0 To matrix.Length - 1
                For j = 0 To matrix(i).Length - 1
                    Console.Write(matrix(i)(j) & "   ")
                Next j
                Console.WriteLine("")
            Next i
        End Sub

        Private Shared Sub ShowMatrix(ByVal matrix()() As Integer, ByVal isTrain As Boolean)
            For i = 0 To matrix.Length - 1
                For j = 0 To matrix(i).Length - 2
                    Console.Write(matrix(i)(j) & " ")
                Next j
                If isTrain = True Then
                    Console.Write(" ->  ")
                    If matrix(i)(matrix(i).Length - 1) = -1 Then
                        Console.Write("-1")
                    Else
                        Console.Write("+1")
                    End If
                End If
                Console.WriteLine("")
            Next i
        End Sub
  End Class ' class Program

  Public Class Learner ' weak learner
	Public feature As Integer ' which feature: Opponent, Field, Spread
	Public value As Integer ' value: Atlanta,.. Home,.. Small,..
	Public predicted As Integer ' -1 (Lose) or +1 (Win)
        Public errorVal As Double ' actual error rate is percentage of tuples incorrectly classified
	Public epsilon As Double ' weighted error (because different training tuples have different weights)
	Public alpha As Double ' importance weight

        Public Sub New(ByVal feature As Integer, ByVal value As Integer, ByVal predicted As Integer, ByVal errorVal As Double, ByVal epsilon As Double, ByVal alpha As Double)
            Me.feature = feature ' ex: "Opponent"
            Me.value = value ' ex: "Chicago"
            Me.predicted = predicted ' -1 or +1 (Lose, Win)
            Me.errorVal = errorVal ' ex: 0.45 (cannot be greater than 0.50)
            Me.epsilon = epsilon ' ex: 0.15 but not 0.65 (i.e., not greater than 0.50)
            Me.alpha = alpha ' ex: 2.36
        End Sub

        Public Overrides Function ToString() As String
            Dim ss As String = ""
            If predicted = 1.0 Then
                ss = "&"
            Else
                ss = ""
            End If
            Dim s As String = ""
            s = s & "feature = " & feature
            s = s & " value = " & value
            s = s & " predicted = " & ss & predicted
            s = s & " errorval = " & errorVal.ToString("F4")
            s = s & " epsilon = " & epsilon.ToString("F4")
            s = s & " alpha = " & alpha.ToString("F4")
            s = s & " alpha = " & alpha.ToString("F4")
            Return s
        End Function
  End Class ' Learner
End Namespace ' ns

